@@ -2,6 +2,7 @@ from time import sleep
 from typing import Final, Type, TypeVar
 
 from serial.serialposix import Serial
+from tabulate import tabulate
 
 CMD_RESET: Final = 0x44
 CMD_STATUS: Final = 0x50
@@ -17,6 +18,12 @@ CMD_GET_BILL_DISPENSE_ORDER: Final = 0x5D
 CMD_SET_BILL_LENGTHS: Final = 0x5E
 CMD_GET_BILL_LENGTHS: Final = 0x5F
 
+NO_ERROR = 0x20
+
+ACK_TIMEOUT = 0xF1
+RESPONSE_TIMEOUT = 0xF2
+UNKNOWN_ERROR = 0xFF
+
 ERROR_CODES = {
     0x01: 'Bill Pick Up Error',
     0x02: 'Jam on the path between CHK Sensor and DVT Sensor',
@@ -41,6 +48,9 @@ ERROR_CODES = {
     0x26: 'Detecting no reject tray before start or for operation',
     0x30: 'Recognizing abnormal command',
     0x31: 'Recognizing abnormal parameter on the command',
+    ACK_TIMEOUT: 'ACK timed out',
+    RESPONSE_TIMEOUT: 'Response timed out',
+    UNKNOWN_ERROR: 'Unknown Error',
 }
 
 
@@ -87,6 +97,7 @@ class ResponseMessage:
     ID: Final[int] = 0x30  # Communications ID
     STX: Final[int] = 0x02  # Start of Text
     RSP: int  # Status Command
+    ERROR: int  # Error Status for Operation
     ...  # Response Parameter (Variable Length)
     ETX: Final[int] = 0x03  # End of Text
     BCC: int  # Block Check Character
@@ -105,6 +116,14 @@ class ResponseMessage:
     def is_valid(self):
         return len(self._origin) > 0 and self.BCC == self._origin[-1]
 
+    def __repr__(self) -> str:
+        return tabulate(
+            ((k, hex(v)) for k, v in self.__dict__.items() if type(v) == int),
+            headers=('Name', 'Code'),
+            colalign=('left', 'right'),
+            tablefmt="pretty"
+        )
+
 
 class ResetCommand(CommandMessage):
     """
@@ -242,7 +261,7 @@ class DispenseCommand(CommandMessage):
     TO1: int = 0x20  # If timeout value isn't used, then 0x20. Else if it's used, the value is 0x1c. Default: 0x20
     TO2: int = 0x20  # If timeout value isn't used, then 0x20. Else if it's used, the value is 0x30~39. Default: 0x20
     RSV: int = 0x20  # Reserved (9 bytes)
-    _response_size = 6 + 2 + 3 * 4 + 1
+    _response_size = 6 + 2 + 3 * 4 + 9
 
     def __init__(self, *args):
         (self.QTY1, self.QTY2, self.QTY3, self.QTY4, self.TO1, self.TO2) = \
@@ -250,10 +269,10 @@ class DispenseCommand(CommandMessage):
         super().__init__(*args)
 
     def _compute_bcc(self, *_):
-        super()._compute_bcc(self.QTY1, self.QTY2, self.QTY3, self.QTY4, self.TO1, self.TO2)
+        super()._compute_bcc(self.QTY1, self.QTY2, self.QTY3, self.QTY4, self.TO1, self.TO2, *[self.RSV] * 9)
 
     def _bytes(self, *_):
-        return super()._bytes(self.QTY1, self.QTY2, self.QTY3, self.QTY4, self.TO1, self.TO2)
+        return super()._bytes(self.QTY1, self.QTY2, self.QTY3, self.QTY4, self.TO1, self.TO2, *[self.RSV] * 9)
 
 
 class DispenseResponse(ResponseMessage):
@@ -281,8 +300,7 @@ class DispenseResponse(ResponseMessage):
             self.EXIT2, self.REJECT2, self.TYPE2,
             self.EXIT3, self.REJECT3, self.TYPE3,
             self.EXIT4, self.REJECT4, self.TYPE4,
-            _
-        ) = response[4:-2]
+        ) = response[4:-2 - 9]
         super().__init__(*response)
 
     def _compute_bcc(self, *_):
@@ -292,7 +310,7 @@ class DispenseResponse(ResponseMessage):
             self.EXIT2, self.REJECT2, self.TYPE2,
             self.EXIT3, self.REJECT3, self.TYPE3,
             self.EXIT4, self.REJECT4, self.TYPE4,
-            self.RSV
+            *[self.RSV] * 9
         )
 
 
@@ -634,11 +652,21 @@ class GetBillLengthsResponse(ResponseMessage):
 
 
 class PuloonException(Exception):
-    message: str
+    code: int = 0x20
+    message: str = ERROR_CODES[UNKNOWN_ERROR]
+    response: ResponseMessage = None
 
-    def __init__(self, message: str, *args: object):
-        super().__init__(*args)
-        self.message = message
+    def __init__(self, code: int, message: str = '', response: ResponseMessage = None, *args: object):
+        self.code = code
+        self.message = message or ERROR_CODES.get(code) or ERROR_CODES[UNKNOWN_ERROR]
+        self.response = response
+        super().__init__(self.code, self.message, self.response, *args)
+
+    def __repr__(self) -> str:
+        _repr = f'{self.__class__.__name__}: {hex(self.code)} - {self.message}'
+        if self.response:
+            _repr += f'\n{repr(self.response)}'
+        return _repr
 
 
 C = TypeVar('C', bound=CommandMessage)
@@ -667,26 +695,25 @@ class PuloonLCDM4000:
             self.port.timeout = self.ack_timeout[1]
 
             ack = self.port.read(1)
-            print(ack)
             if len(ack) > 0 and self.ack == ack[0]:
                 break
         else:
-            raise PuloonException('ACK timed out')
+            raise PuloonException(ACK_TIMEOUT)
 
         for _ in range(3):
             self.port.timeout = self.response_delay[1]
             data = self.port.read(cmd.response_size())
-            print(data)
             resp = resp_class(*data)
             if resp.is_valid():
                 self.port.write(chr(self.ack).encode('ascii'))
                 eot = self.port.read(1)
-                print(eot)
                 if len(eot) > 0 and self.eot == eot[0]:
+                    if resp.ERROR != NO_ERROR:
+                        raise PuloonException(resp.ERROR - NO_ERROR, response=resp)
                     break
             self.port.write(chr(self.nak).encode('ascii'))
         else:
-            raise PuloonException('Response timed out')
+            raise PuloonException(RESPONSE_TIMEOUT)
 
         return resp
 
@@ -709,11 +736,17 @@ class PuloonLCDM4000:
         return self._run(DispenseCommand, DispenseResponse, *args, to1, to2)
 
     def test_dispense(self, qty1: int = 0, qty2: int = 0, qty3: int = 0, qty4: int = 0,
-                      to1: int = 0x20, to2: int = 0x20) -> TestDispenseResponse:
+                      to1: bool = False, to2: int = 0) -> TestDispenseResponse:
         assert all(0 <= q <= 40 for q in (qty1, qty2, qty3, qty4)), "Quantity should be in range of 0 and 40"
         assert 0 < sum((qty1, qty2, qty3, qty4)) <= 40, "Quantity should be at least 1 and maximum 40 in total"
-        args = [q + 0x20 for q in (qty1, qty2, qty3, qty4)]
-        return self._run(TestDispenseCommand, TestDispenseResponse, *args, to1, to2)
+        assert 0 <= to2 <= 9, "Timeout should be in range of 0 and 9"
+        return self._run(
+            TestDispenseCommand,
+            TestDispenseResponse,
+            *(q + 0x20 for q in (qty1, qty2, qty3, qty4)),
+            0x1C if to1 else 0x20,
+            to2 + 0x30 if to1 else 0x20
+        )
 
     def last_status(self) -> LastStatusResponse:
         return self._run(LastStatusCommand, LastStatusResponse)
